/* 
 * HSG Audit VFS module. Logs noteworthy events via samba log mechanisms
 *
 * Copyright (C) Tim Potter, 1999-2000
 * Copyright (C) Alexander Bokovoy, 2002
 * Copyright (C) Stefan (metze) Metzmacher, 2003
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *  
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *  
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include "includes.h"

static void hsgaudit_log(vfs_handle_struct *handle, connection_struct *conn, const char *msg)
{
	DEBUG(1, ("AUDIT: %s@%s: %s\n", conn->user, conn->client_address, msg));
}

static int hsgaudit_connect(vfs_handle_struct *handle, connection_struct *conn, const char *service, const char *user)    
{
	pstring msg;
	
	snprintf(msg, sizeof(msg)-1, "connect to service %s", service);
	hsgaudit_log(handle, conn, msg);
	return SMB_VFS_NEXT_CONNECT(handle, conn, service, user);
}

static void hsgaudit_disconnect(vfs_handle_struct *handle, connection_struct *conn)
{
	pstring msg;
	
	snprintf(msg, sizeof(msg)-1, "disconnected");
	hsgaudit_log(handle, conn, msg);
	SMB_VFS_NEXT_DISCONNECT(handle, conn);
}

static int hsgaudit_mkdir(vfs_handle_struct *handle, connection_struct *conn, const char *path, mode_t mode)
{
	int result;
	pstring msg;
	
	result = SMB_VFS_NEXT_MKDIR(handle, conn, path, mode);
	snprintf(msg, sizeof(msg)-1, "mkdir %s%s%s", path, (result < 0) ? " failed: " : "", (result < 0) ? strerror(errno) : "");
	hsgaudit_log(handle, conn, msg);
	return result;
}

static int hsgaudit_rmdir(vfs_handle_struct *handle, connection_struct *conn, const char *path)
{
	int result;
	pstring msg;
	
	result = SMB_VFS_NEXT_RMDIR(handle, conn, path);
	snprintf(msg, sizeof(msg)-1, "rmdir %s%s%s", path, (result < 0) ? " failed: " : "", (result < 0) ? strerror(errno) : "");
	hsgaudit_log(handle, conn, msg);
	return result;
}

static int hsgaudit_rename(vfs_handle_struct *handle, connection_struct *conn, const char *old, const char *new)
{
	int result;
	pstring msg;
	
	result = SMB_VFS_NEXT_RENAME(handle, conn, old, new);
	snprintf(msg, sizeof(msg)-1, "rename %s -> %s%s%s", old, new, (result < 0) ? " failed: " : "", (result < 0) ? strerror(errno) : "");
	hsgaudit_log(handle, conn, msg);
	return result;
}

static int hsgaudit_unlink(vfs_handle_struct *handle, connection_struct *conn, const char *path)
{
	int result;
	pstring msg;
	
	result = SMB_VFS_NEXT_UNLINK(handle, conn, path);
	snprintf(msg, sizeof(msg)-1, "unlink %s%s%s", path, (result < 0) ? " failed: " : "", (result < 0) ? strerror(errno) : "");
	hsgaudit_log(handle, conn, msg);
	return result;
}

/* VFS operations structure */

static vfs_op_tuple hsgaudit_op_tuples[] = {

	/* Disk operations */

	{SMB_VFS_OP(hsgaudit_connect),		SMB_VFS_OP_CONNECT, 	SMB_VFS_LAYER_TRANSPARENT},
	{SMB_VFS_OP(hsgaudit_disconnect),	SMB_VFS_OP_DISCONNECT,	SMB_VFS_LAYER_TRANSPARENT},
	
	/* Directory operations */

	{SMB_VFS_OP(hsgaudit_mkdir),		SMB_VFS_OP_MKDIR,	SMB_VFS_LAYER_TRANSPARENT},
	{SMB_VFS_OP(hsgaudit_rmdir),		SMB_VFS_OP_RMDIR,	SMB_VFS_LAYER_TRANSPARENT},

	/* File operations */

	{SMB_VFS_OP(hsgaudit_rename),		SMB_VFS_OP_RENAME,	SMB_VFS_LAYER_TRANSPARENT},
	{SMB_VFS_OP(hsgaudit_unlink),		SMB_VFS_OP_UNLINK,	SMB_VFS_LAYER_TRANSPARENT},

	{NULL,					SMB_VFS_OP_NOOP,	SMB_VFS_LAYER_NOOP}
};

NTSTATUS init_module(void)
{
	return smb_register_vfs(SMB_VFS_INTERFACE_VERSION, "hsgaudit", hsgaudit_op_tuples);
}
