/* 
 * Roaming Profile VFS module.  Renames NTUSER.DAT to NTUSER.MAN transparently 
 * for all user expect the Profile Administrator
 *
 * Copyright (C) Tim Potter, 1999-2000
 * Copyright (C) Alexander Bokovoy, 2002
 * Copyright (C) Stefan (metze) Metzmacher, 2003
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *  
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *  
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include "includes.h"

#define NTUSER_DEBUG (3)

/* MUST have same length and same position of case-sensitive chars */
static const char *virt = "NTUSER.MAN";
static const char *real = "NTUSER.DAT";

/* support functions */

#define REMAP(hnd) ((int)hnd->data)

static void noop(void **data)
{
}

static void strncasepreservecpy(char *dst, const char *src)
{
	for (; *dst && *src; dst++, src++) {
		if (isupper(*dst))
			*dst = toupper(*src);
		else if (islower(*dst))
			*dst = tolower(*src);
		else *dst = *src;
	}
}

static __inline__ void test_x_and_replace_y(char *str, const char *x, const char *y)
{
	char *p;
	
	while ((p = strcasestr(str, x)) != NULL)
		strncasepreservecpy(p, y);
}

static __inline__ void test_virt_and_replace_real(char *str)
{
	test_x_and_replace_y(str, virt, real);
}

static __inline__ void test_real_and_replace_virt(char *str)
{
	test_x_and_replace_y(str, real, virt);
}

static __inline__ const char *get_realname(vfs_handle_struct *handle, connection_struct *conn, const char *fname, char *realname, size_t len)
{
	if (REMAP(handle) && fname) {
		strncpy(realname, fname, len);
		if (!strcasestr(fname, real))
			test_virt_and_replace_real(realname);
		else
			test_real_and_replace_virt(realname); /* remapping in effect and real file name requested => make invalid */
		return realname;
	} else
		return fname;
}

#define GET_REALNAME(handle, conn, fname, realname, name) \
	const char *name; \
	pstring realname; \
	name = get_realname(handle, conn, fname, realname, sizeof(realname)-1); \
	DEBUG(NTUSER_DEBUG, ("%s(..., \"%s\" => \"%s\", ...)\n", __FUNCTION__, fname, name))

/* VFS functions */

static int ntuser_connect(vfs_handle_struct *handle, connection_struct *conn, const char *service, const char *user)    
{
	REMAP(handle) = !user || !handle || !handle->param || strcmp(user, handle->param);
	handle->free_data = noop;
	
	DEBUG(NTUSER_DEBUG, ("%s(0x%p (param=\"%s\"), 0x%p, \"%s\", \"%s\") => %sremapping\n",
		__FUNCTION__, handle, handle->param, conn, service, user, REMAP(handle) ? "" : "not "));
	
	return SMB_VFS_NEXT_CONNECT(handle, conn, service, user);
}

static struct dirent *ntuser_readdir(vfs_handle_struct *handle, connection_struct *conn, DIR *dirp)
{
	struct dirent *dent;
	
	dent = SMB_VFS_NEXT_READDIR(handle, conn, dirp);
	if (REMAP(handle) && dent)
		test_real_and_replace_virt(dent->d_name);
	DEBUG(NTUSER_DEBUG, ("%s(...) => %s\n", __FUNCTION__, dent ? dent->d_name : NULL));
	
	return dent;
}

static int ntuser_open(vfs_handle_struct *handle, connection_struct *conn, const char *fname, int flags, mode_t mode)
{
	GET_REALNAME(handle, conn, fname, realname, name);
	return SMB_VFS_NEXT_OPEN(handle, conn, name, flags, mode);
}

static int ntuser_stat(vfs_handle_struct *handle, connection_struct *conn, const char *fname, SMB_STRUCT_STAT *sbuf)
{
	GET_REALNAME(handle, conn, fname, realname, name);
	return SMB_VFS_NEXT_STAT(handle, conn, name, sbuf);
}

static int ntuser_lstat(vfs_handle_struct *handle, connection_struct *conn, const char *path, SMB_STRUCT_STAT *sbuf)
{
	GET_REALNAME(handle, conn, path, realname, name);
	return SMB_VFS_NEXT_LSTAT(handle, conn, name, sbuf);
}

static int ntuser_utime(vfs_handle_struct *handle, connection_struct *conn, const char *path, struct utimbuf *times)
{
	GET_REALNAME(handle, conn, path, realname, name);
	return SMB_VFS_NEXT_UTIME(handle, conn, name, times);
}

static BOOL ntuser_readlink(vfs_handle_struct *handle, connection_struct *conn, const char *path, char *buf, size_t bufsiz)
{
	GET_REALNAME(handle, conn, path, realname, name);
	return SMB_VFS_NEXT_READLINK(handle, conn, name, buf, bufsiz);
}

static size_t ntuser_get_nt_acl(vfs_handle_struct *handle, files_struct *fsp, const char *path, uint32 security_info, struct security_descriptor_info **ppdesc)
{
	GET_REALNAME(handle, NULL, path, realname, name);
	return SMB_VFS_NEXT_GET_NT_ACL(handle, fsp, name, security_info, ppdesc);
}

static SMB_ACL_T ntuser_sys_acl_get_file(vfs_handle_struct *handle, connection_struct *conn, const char *path_p, SMB_ACL_TYPE_T type)
{
	GET_REALNAME(handle, conn, path_p, realname, name);
	return SMB_VFS_NEXT_SYS_ACL_GET_FILE(handle, conn, name, type);
}

static ssize_t ntuser_getxattr(vfs_handle_struct *handle, struct connection_struct *conn,const char *path, const char *name, void *value, size_t size)
{
	GET_REALNAME(handle, conn, path, realname, name_p);
        return SMB_VFS_NEXT_GETXATTR(handle, conn, name_p, name, value, size);
}

static ssize_t ntuser_lgetxattr(vfs_handle_struct *handle, struct connection_struct *conn,const char *path, const char *name, void *value, size_t
size)
{
	GET_REALNAME(handle, conn, path, realname, name_p);
        return SMB_VFS_NEXT_LGETXATTR(handle, conn, name_p, name, value, size);
}

static ssize_t ntuser_listxattr(vfs_handle_struct *handle, struct connection_struct *conn,const char *path, char *list, size_t size)
{
	GET_REALNAME(handle, conn, path, realname, name);
        return SMB_VFS_NEXT_LISTXATTR(handle, conn, name, list, size);
}

static ssize_t ntuser_llistxattr(vfs_handle_struct *handle, struct connection_struct *conn,const char *path, char *list, size_t size)
{
	GET_REALNAME(handle, conn, path, realname, name);
        return SMB_VFS_NEXT_LLISTXATTR(handle, conn, name, list, size);
}

/* VFS operations structure */

static vfs_op_tuple ntuser_op_tuples[] = {

	/* Disk operations */

	{SMB_VFS_OP(ntuser_connect),		SMB_VFS_OP_CONNECT, 		SMB_VFS_LAYER_TRANSPARENT},
	
	/* Directory operations */

	{SMB_VFS_OP(ntuser_readdir),		SMB_VFS_OP_READDIR,		SMB_VFS_LAYER_TRANSPARENT},

	/* File operations */

	{SMB_VFS_OP(ntuser_open),		SMB_VFS_OP_OPEN,		SMB_VFS_LAYER_TRANSPARENT},
	{SMB_VFS_OP(ntuser_stat),		SMB_VFS_OP_STAT,		SMB_VFS_LAYER_TRANSPARENT},
	{SMB_VFS_OP(ntuser_lstat),		SMB_VFS_OP_LSTAT,		SMB_VFS_LAYER_TRANSPARENT},
	{SMB_VFS_OP(ntuser_utime),		SMB_VFS_OP_UTIME,		SMB_VFS_LAYER_TRANSPARENT},
	{SMB_VFS_OP(ntuser_readlink),		SMB_VFS_OP_READLINK,		SMB_VFS_LAYER_TRANSPARENT},

	/* NT File ACL operations */

	{SMB_VFS_OP(ntuser_get_nt_acl),		SMB_VFS_OP_GET_NT_ACL,		SMB_VFS_LAYER_TRANSPARENT},

	/* POSIX ACL operations */

	{SMB_VFS_OP(ntuser_sys_acl_get_file),	SMB_VFS_OP_SYS_ACL_GET_FILE,	SMB_VFS_LAYER_TRANSPARENT},
	
	/* EA operations. */
	{SMB_VFS_OP(ntuser_getxattr),		SMB_VFS_OP_GETXATTR,		SMB_VFS_LAYER_TRANSPARENT},
	{SMB_VFS_OP(ntuser_lgetxattr),		SMB_VFS_OP_LGETXATTR,		SMB_VFS_LAYER_TRANSPARENT},
	{SMB_VFS_OP(ntuser_listxattr),		SMB_VFS_OP_LISTXATTR,		SMB_VFS_LAYER_TRANSPARENT},
	{SMB_VFS_OP(ntuser_llistxattr),		SMB_VFS_OP_LLISTXATTR,		SMB_VFS_LAYER_TRANSPARENT},

	{NULL,					SMB_VFS_OP_NOOP,		SMB_VFS_LAYER_NOOP}
};

NTSTATUS init_module(void)
{
	return smb_register_vfs(SMB_VFS_INTERFACE_VERSION, "ntuser", ntuser_op_tuples);
}

