/*
 * pgSQL password backend for samba
 * Copyright (C) Maurice Massar 2003
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 675
 * Mass Ave, Cambridge, MA 02139, USA.
 */


#include "includes.h"
#include <libpq-fe.h>

static int pgsqlsam_debug_level = DBGC_ALL;

#undef DBGC_CLASS
#define DBGC_CLASS pgsqlsam_debug_level

typedef struct {
	PGconn *conn;
	char *domain;
	const char *location;
} pgsqlam_private;

static void free_private_data(void **vp)
{
	pgsqlam_private **prv = (pgsqlam_private **)vp;
	
	PQfinish((*prv)->conn);
	DEBUG(5, ("Disconnected from postgresql db\n"));
	(*prv)->conn = NULL;
	/* No need to free any further, as it is talloc()ed */
}

typedef enum {
	LOGON_TIME, /* column first */
	LOGOFF_TIME,
	KICKOFF_TIME,
	PASS_LAST_SET_TIME,
	PASS_CAN_CHANGE_TIME,
	PASS_MUST_CHANGE_TIME,
	USERNAME,
	DOMAIN,
	NT_USERNAME,
	FULLNAME,
	HOME_DIR,
	DIR_DRIVE,
	LOGON_SCRIPT,
	PROFILE_PATH,
	ACCT_DESC,
	WORKSTATIONS,
	MUNGED_DIAL,
	USER_SID,
	GROUP_SID,
	LM_PW,
	NT_PW,
	ACCT_CTRL,
	LOGON_DIVS,
	HOURS_LEN,
	HOURS,
	ATTR_COUNT, /**/
	TABLE,
	CONNECTINFO,
	CONFIG_ALL
} pgsql_config;

const char *config_default[] = {
	"logon_time",
	"logoff_time",
	"kickoff_time",
	"pass_last_set_time",
	"pass_can_change_time",
	"pass_must_change_time",
	"username",
	"domain",
	"nt_username",
	"fullname",
	"home_dir",
	"dir_drive",
	"logon_script",
	"profile_path",
	"acct_desc",
	"workstations",
	"munged_dial",
	"user_sid",
	"group_sid",
	"lm_pw",
	"nt_pw",
	"acct_ctrl",
	"logon_divs",
	"hours_len",
	"hours",
	"-- attr count --",
	"user",
	"",
	"-- config all --"
};

const char *config_key[] = {
	"logon time column",
	"logoff time column",
	"kickoff time column",
	"pass last set time column",
	"pass can change time column",
	"pass must change time column",
	"username column",
	"domain column",
	"nt username column",
	"fullname column",
	"home dir column",
	"dir drive column",
	"logon script column",
	"profile path column",
	"acct desc column",
	"workstations column",
	"munged dial column",
	"user sid column",
	"group sid column",
	"lm pw column",
	"nt pw column",
	"acct ctrl column",
	"logon divs column",
	"hours len column",
	"hours column",
	"-- attr count --",
	"table",
	"connect info",
	"-- config all --"
};

#define config_value(prv, name) \
	lp_parm_const_string(GLOBAL_SECTION_SNUM, (prv)->location, config_key[name], config_default[name])

static NTSTATUS res_to_sam(pgsqlam_private *prv, PGresult *res, SAM_ACCOUNT *user)
{
	int col;
	char *tmp;
	const char *ctmp;
	time_t tmpt;
	int tmpi;
	uint8 hours[MAX_HOURS_LEN];
	uint8 smblmpwd[LM_HASH_LEN], smbntpwd[NT_HASH_LEN];
	struct passwd *pw = NULL;
	uid_t uid = -1;
	gid_t gid = -1;
	
	if (PQntuples(res) != 1)
		return NT_STATUS_UNSUCCESSFUL;
	
	if (((col = PQfnumber(res, config_value(prv, USERNAME))) == -1)||PQgetisnull(res, 0, col)) {
		DEBUG(0, ("NULL username found!\n"));
		return NT_STATUS_UNSUCCESSFUL;
	} else {
		tmp = PQgetvalue(res, 0, PQfnumber(res, config_value(prv, USERNAME)));
		pdb_set_username(user, tmp, PDB_SET);
		if ((pw = Get_Pwnam(tmp)) != NULL) {
			uid = pw->pw_uid;
			gid = pw->pw_gid;
		}
	}
	
	if (((col = PQfnumber(res, config_value(prv, DOMAIN))) == -1)||PQgetisnull(res, 0, col)) {
		pdb_set_domain(user, prv->domain, PDB_DEFAULT);
	} else {
		tmp = PQgetvalue(res, 0, col);
		pdb_set_username(user, tmp, PDB_SET);
	}
	
	if (((col = PQfnumber(res, config_value(prv, NT_USERNAME))) == -1)||PQgetisnull(res, 0, col)) {
		ctmp = pdb_get_username(user);
		pdb_set_nt_username(user, ctmp, PDB_DEFAULT);
	} else {
		tmp = PQgetvalue(res, 0, col);
		pdb_set_nt_username(user, tmp, PDB_SET);
	}
	
	if (((col = PQfnumber(res, config_value(prv, USER_SID))) == -1)||PQgetisnull(res, 0, col)) {
		/* leave default */
	} else {
		tmp = PQgetvalue(res, 0, col);
		pdb_set_user_sid_from_string(user, tmp, PDB_SET);
	}
	
	if (((col = PQfnumber(res, config_value(prv, GROUP_SID))) == -1)||PQgetisnull(res, 0, col)) {
		pdb_set_group_sid_from_rid(user, DOMAIN_GROUP_RID_USERS, PDB_DEFAULT);
	} else {
		tmp = PQgetvalue(res, 0, col);
		pdb_set_group_sid_from_string(user, tmp, PDB_SET);
	}
	
	if (((col = PQfnumber(res, config_value(prv, PASS_LAST_SET_TIME))) == -1)||PQgetisnull(res, 0, col)) {
		/* leave default */
	} else {
		tmpt = (time_t) atol(PQgetvalue(res, 0, col));
		pdb_set_pass_last_set_time(user, tmpi, PDB_SET);
	}
	
	if (((col = PQfnumber(res, config_value(prv, PASS_CAN_CHANGE_TIME))) == -1)||PQgetisnull(res, 0, col)) {
		/* leave default */
	} else {
		tmpt = (time_t) atol(PQgetvalue(res, 0, col));
		pdb_set_pass_can_change_time(user, tmpi, PDB_SET);
	}
	
	if (((col = PQfnumber(res, config_value(prv, PASS_MUST_CHANGE_TIME))) == -1)||PQgetisnull(res, 0, col)) {
		/* leave default */
	} else {
		tmpt = (time_t) atol(PQgetvalue(res, 0, col));
		pdb_set_pass_must_change_time(user, tmpi, PDB_SET);
	}
	
	if (((col = PQfnumber(res, config_value(prv, LOGON_TIME))) == -1)||PQgetisnull(res, 0, col)) {
		/* leave default */
	} else {
		tmpt = (time_t) atol(PQgetvalue(res, 0, col));
		pdb_set_logon_time(user, tmpi, PDB_SET);
	}
	
	if (((col = PQfnumber(res, config_value(prv, LOGOFF_TIME))) == -1)||PQgetisnull(res, 0, col)) {
		/* leave default */
	} else {
		tmpt = (time_t) atol(PQgetvalue(res, 0, col));
		pdb_set_logoff_time(user, tmpi, PDB_SET);
	}
	
	if (((col = PQfnumber(res, config_value(prv, KICKOFF_TIME))) == -1)||PQgetisnull(res, 0, col)) {
		/* leave default */
	} else {
		tmpt = (time_t) atol(PQgetvalue(res, 0, col));
		pdb_set_kickoff_time(user, tmpi, PDB_SET);
	}
	
	if (((col = PQfnumber(res, config_value(prv, FULLNAME))) == -1)||PQgetisnull(res, 0, col)) {
		/* leave default */
	} else {
		tmp = PQgetvalue(res, 0, col);
		pdb_set_fullname(user, tmp, PDB_SET);
	}
	
	if (((col = PQfnumber(res, config_value(prv, ACCT_DESC))) == -1)||PQgetisnull(res, 0, col)) {
		/* leave default */
	} else {
		tmp = PQgetvalue(res, 0, col);
		pdb_set_acct_desc(user, tmp, PDB_SET);
	}
	
	if (((col = PQfnumber(res, config_value(prv, WORKSTATIONS))) == -1)||PQgetisnull(res, 0, col)) {
		/* leave default */
	} else {
		tmp = PQgetvalue(res, 0, col);
		pdb_set_workstations(user, tmp, PDB_SET);
	}
	
	if (((col = PQfnumber(res, config_value(prv, MUNGED_DIAL))) == -1)||PQgetisnull(res, 0, col)) {
		/* leave default */
	} else {
		tmp = PQgetvalue(res, 0, col);
		pdb_set_munged_dial(user, tmp, PDB_SET);
	}
	
	if (((col = PQfnumber(res, config_value(prv, LOGON_DIVS))) == -1)||PQgetisnull(res, 0, col)) {
		pdb_set_logon_divs(user, 168, PDB_DEFAULT);
	} else {
		tmpi = atol(PQgetvalue(res, 0, col));
		pdb_set_logon_divs(user, tmpi, PDB_SET);
	}
	
	if (((col = PQfnumber(res, config_value(prv, HOURS_LEN))) == -1)||PQgetisnull(res, 0, col)) {
		pdb_set_logon_divs(user, 21, PDB_DEFAULT);
	} else {
		tmpi = atol(PQgetvalue(res, 0, col));
		pdb_set_logon_divs(user, tmpi, PDB_SET);
	}
	
	if (((col = PQfnumber(res, config_value(prv, HOURS))) == -1)||PQgetisnull(res, 0, col)) {
		memset(hours, 0xff, MAX_HOURS_LEN);
		pdb_set_hours(user, hours, PDB_DEFAULT);
	} else {
		tmp = PQgetvalue(res, 0, col);
		memset(hours, 0xff, MAX_HOURS_LEN); /*TODO: convert tmp to hours */
		pdb_set_hours(user, hours, PDB_SET);
	}
	
	if (((col = PQfnumber(res, config_value(prv, ACCT_CTRL))) == -1)||PQgetisnull(res, 0, col)) {
		pdb_set_acct_ctrl(user, ACB_NORMAL, PDB_DEFAULT);
	} else {
		tmp = PQgetvalue(res, 0, col);
		tmpi = pdb_decode_acct_ctrl(tmp);
		if (tmpi == 0)
			tmpi = ACB_NORMAL;
		pdb_set_acct_ctrl(user, tmpi, PDB_SET);
	}
	
	if (((col = PQfnumber(res, config_value(prv, LM_PW))) == -1)||PQgetisnull(res, 0, col)) {
		/* leave default */
	} else {
		tmp = PQgetvalue(res, 0, col);
		pdb_gethexpwd(tmp, smblmpwd);
		memset(tmp, 0, strlen(tmp)+1);
		pdb_set_lanman_passwd(user, smblmpwd, PDB_SET);
		ZERO_STRUCT(smblmpwd);
	}
	
	if (((col = PQfnumber(res, config_value(prv, NT_PW))) == -1)||PQgetisnull(res, 0, col)) {
		/* leave default */
	} else {
		tmp = PQgetvalue(res, 0, col);
		pdb_gethexpwd(tmp, smbntpwd);
		memset(tmp, 0, strlen(tmp)+1);
		pdb_set_lanman_passwd(user, smbntpwd, PDB_SET);
		ZERO_STRUCT(smbntpwd);
	}
	
	if (((col = PQfnumber(res, config_value(prv, DIR_DRIVE))) == -1)||PQgetisnull(res, 0, col)) {
		pdb_set_dir_drive(user,
			talloc_sub_specified(user->mem_ctx, lp_logon_drive(),
				pdb_get_username(user), pdb_get_domain(user),
				uid, gid),
			PDB_DEFAULT);
	} else {
		tmp = PQgetvalue(res, 0, col);
		pdb_set_dir_drive(user, tmp, PDB_SET);
	}
	
	if (((col = PQfnumber(res, config_value(prv, HOME_DIR))) == -1)||PQgetisnull(res, 0, col)) {
		pdb_set_homedir(user,
			talloc_sub_specified(user->mem_ctx, lp_logon_home(),
				pdb_get_username(user), pdb_get_domain(user),
				uid, gid),
			PDB_DEFAULT);
	} else {
		tmp = PQgetvalue(res, 0, col);
		pdb_set_homedir(user, tmp, PDB_SET);
	}
	
	if (((col = PQfnumber(res, config_value(prv, LOGON_SCRIPT))) == -1)||PQgetisnull(res, 0, col)) {
		pdb_set_logon_script(user,
			talloc_sub_specified(user->mem_ctx, lp_logon_script(),
				pdb_get_username(user), pdb_get_domain(user),
				uid, gid),
			PDB_DEFAULT);
	} else {
		tmp = PQgetvalue(res, 0, col);
		pdb_set_logon_script(user, tmp, PDB_SET);
	}
	
	if (((col = PQfnumber(res, config_value(prv, PROFILE_PATH))) == -1)||PQgetisnull(res, 0, col)) {
		pdb_set_profile_path(user,
			talloc_sub_specified(user->mem_ctx, lp_logon_path(),
				pdb_get_username(user), pdb_get_domain(user),
				uid, gid),
			PDB_DEFAULT);
	} else {
		tmp = PQgetvalue(res, 0, col);
		pdb_set_profile_path(user, tmp, PDB_SET);
	}
	
	return NT_STATUS_OK;
}

static void dopqclear(pgsqlam_private *prv, PGresult *res)
{
	PQclear(res);
}

static PGresult *dopqexec(pgsqlam_private *prv, const char *query, int need_res)
{
	PGresult *res;
	
	res = PQexec(prv->conn, query);
	if (need_res)
		return res;
	
	dopqclear(prv, res);
}

static char *dopqesacpe(pgsqlam_private *prv, const char *val)
{
	size_t len;
	char *sql_val;
	
	len = strlen(val);
	sql_val = malloc(len * 2 + 1);
	PQescapeString(sql_val, val, len);
	
	return sql_val;
}

static NTSTATUS res_to_status(pgsqlam_private *prv, PGresult *res)
{
	/*TODO*/
	return NT_STATUS_OK;
}

/***************************************************************
 Start enumeration of the passwd list.
****************************************************************/

static NTSTATUS pgsqlsam_setsampwent(struct pdb_methods *methods, BOOL update)
{
	char *query;
	pgsqlam_private *prv = (pgsqlam_private *) methods->private_data;
	
	DEBUG(10, ("%s(pdbm-%s, update=%d)\n", __FUNCTION__, methods->name, update));
	
	if (update)
		return NT_STATUS_NOT_IMPLEMENTED;
	
	asprintf(&query, "BEGIN; DECLARE userc CURSOR FOR SELECT * FROM %s FOR READ ONLY;", config_value(prv, TABLE));
	dopqexec(prv, query, 0);
	SAFE_FREE(query);
	
	return NT_STATUS_OK;
}

/***************************************************************
 End enumeration of the passwd list.
****************************************************************/

static void pgsqlsam_endsampwent(struct pdb_methods *methods)
{
	DEBUG(10, ("%s(pdbm-%s)\n", __FUNCTION__, methods->name));
	pgsqlam_private *prv = (pgsqlam_private *) methods->private_data;
	
	dopqexec(prv, "CLOSE userc; COMMIT;", 0);
}

/*****************************************************************
 Get one SAM_ACCOUNT from the list (next in line)
*****************************************************************/

static NTSTATUS pgsqlsam_getsampwent(struct pdb_methods *methods, SAM_ACCOUNT *user)
{
	NTSTATUS ret;
	PGresult *res;
	pgsqlam_private *prv = (pgsqlam_private *) methods->private_data;
	
	/*DEBUG(10, ("%s(pdbm-%s, (out))\n", __FUNCTION__, methods->name));*/
	
	res = dopqexec(prv, "FETCH 1 FROM userc;", 1);
	ret = res_to_sam(prv, res, user);
	dopqclear(prv, res);
	
	return ret;
}

/******************************************************************
 Lookup in the SAM database
******************************************************************/

static NTSTATUS getsam_by_col (struct pdb_methods *methods, SAM_ACCOUNT *user, const char *attr, const char *val)
{
	NTSTATUS ret;
	char *query, *sql_val;
	PGresult *res;
	pgsqlam_private *prv = (pgsqlam_private *) methods->private_data;
	
	DEBUG(10, ("%s(pdbm-%s, (out), \"%s\", \"%s\")\n", __FUNCTION__, methods->name, attr, val));
	
	sql_val = dopqesacpe(prv, val);
	asprintf(&query, "SELECT * FROM %s WHERE %s = '%s';", config_value(prv, TABLE), attr, sql_val);
	res = dopqexec(prv, query, 1);
	SAFE_FREE(query);
	SAFE_FREE(sql_val);
	ret = res_to_sam(prv, res, user);
	dopqclear(prv, res);
	
	return ret;
}

/***************************************************************************
 Search by name
 **************************************************************************/

static NTSTATUS pgsqlsam_getsampwnam (struct pdb_methods *methods, SAM_ACCOUNT *user, const char *sname)
{
	pgsqlam_private *prv = (pgsqlam_private *) methods->private_data;
	DEBUG(10, ("%s(pdbm-%s, (out), \"%s\")\n", __FUNCTION__, methods->name, sname));
	return getsam_by_col(methods, user, config_value(prv, USERNAME), sname);
}

/***************************************************************************
 Search by sid
 **************************************************************************/

static NTSTATUS pgsqlsam_getsampwsid (struct pdb_methods *methods, SAM_ACCOUNT *user, const DOM_SID *sid)
{
	pstring sid_string;
	pgsqlam_private *prv = (pgsqlam_private *) methods->private_data;
	DEBUG(10, ("%s(pdbm-%s, (out), %s)\n", __FUNCTION__, methods->name, sid_to_string(sid_string, sid)));
	return getsam_by_col(methods, user, config_value(prv, USER_SID), sid_to_string(sid_string, sid));
}

/***************************************************************************
 Delete a SAM_ACCOUNT
****************************************************************************/

static NTSTATUS pgsqlsam_delete_sam_account(struct pdb_methods *methods, SAM_ACCOUNT *sam_pass)
{
	NTSTATUS ret;
	char *query, *sql_user;
	PGresult *res;
	pgsqlam_private *prv = (pgsqlam_private *) methods->private_data;
	
	DEBUG(10, ("%s(pdbm-%s, \"%s\")\n", __FUNCTION__, methods->name, pdb_get_username(sam_pass)));
	
	sql_user = dopqesacpe(prv, pdb_get_username(sam_pass));
	asprintf(&query, "DELETE FROM %s WHERE %s = '%s';",
		config_value(prv, TABLE),
		config_value(prv, USERNAME),
		sql_user);
	res = dopqexec(prv, query, 1);
	SAFE_FREE(query);
	SAFE_FREE(sql_user);
	ret = res_to_status(prv, res);
	dopqclear(prv, res);
	
	return ret;
}

/***************************************************************************
 Modifies an existing SAM_ACCOUNT
****************************************************************************/

static NTSTATUS pgsqlsam_update_sam_account (struct pdb_methods *methods, SAM_ACCOUNT *newpwd)
{
	DEBUG(10, ("%s(pdbm-%s, \"%s\")\n", __FUNCTION__, methods->name, pdb_get_username(newpwd)));
	return NT_STATUS_NOT_IMPLEMENTED;
}

/***************************************************************************
 Adds an existing SAM_ACCOUNT
****************************************************************************/

static NTSTATUS pgsqlsam_add_sam_account (struct pdb_methods *methods, SAM_ACCOUNT *newpwd)
{
	DEBUG(10, ("%s(pdbm-%s, \"%s\")\n", __FUNCTION__, methods->name, pdb_get_username(newpwd)));
	return NT_STATUS_NOT_IMPLEMENTED;
}

/***************************************************************************
 Search group by sid
 **************************************************************************/

static NTSTATUS pgsqlsam_getgrsid(struct pdb_methods *methods, GROUP_MAP *map, DOM_SID sid)
{
	pstring sid_string;
	DEBUG(10, ("%s(pdbm-%s, (out), %s)\n", __FUNCTION__, methods->name, sid_to_string(sid_string, &sid)));
	return NT_STATUS_NOT_IMPLEMENTED;
}

/***************************************************************************
 Search group by gid
 **************************************************************************/

static NTSTATUS pgsqlsam_getgrgid(struct pdb_methods *methods, GROUP_MAP *map, gid_t gid)
{
	DEBUG(10, ("%s(pdbm-%s, (out), %d)\n", __FUNCTION__, methods->name, gid));
	return NT_STATUS_NOT_IMPLEMENTED;
}

/***************************************************************************
 Search group by name (unix or windows name)
 **************************************************************************/

static NTSTATUS pgsqlsam_getgrnam(struct pdb_methods *methods, GROUP_MAP *map, const char *name)
{
	DEBUG(10, ("%s(pdbm-%s, (out), \"%s\")\n", __FUNCTION__, methods->name, name));
	return NT_STATUS_NOT_IMPLEMENTED;
}

/***************************************************************************
 Add a group mapping
 **************************************************************************/

static NTSTATUS pgsqlsam_add_group_mapping_entry(struct pdb_methods *methods, GROUP_MAP *map)
{
	pstring sid_string;
	DEBUG(10, ("%s(pdbm-%s, (%d<->%s \"%s\" (%d)))\n", __FUNCTION__, methods->name,
		map->gid, sid_to_string(sid_string, &(map->sid)), map->nt_name, map->sid_name_use));
	return NT_STATUS_NOT_IMPLEMENTED;
}

/***************************************************************************
 Update a group mapping
 **************************************************************************/

static NTSTATUS pgsqlsam_update_group_mapping_entry(struct pdb_methods *methods, GROUP_MAP *map)
{
	pstring sid_string;
	DEBUG(10, ("%s(pdbm-%s, (%d<->%s \"%s\" (%d)))\n", __FUNCTION__, methods->name,
		map->gid, sid_to_string(sid_string, &(map->sid)), map->nt_name, map->sid_name_use));
	DEBUG(10, ("%s(pdbm-%s)\n", __FUNCTION__, methods->name));
	return NT_STATUS_NOT_IMPLEMENTED;
}

/***************************************************************************
 Delete a group mapping
 **************************************************************************/

static NTSTATUS pgsqlsam_delete_group_mapping_entry(struct pdb_methods *methods, DOM_SID sid)
{
	pstring sid_string;
	DEBUG(10, ("%s(pdbm-%s, %s)\n", __FUNCTION__, methods->name, sid_to_string(sid_string, &sid)));
	DEBUG(10, ("%s(pdbm-%s)\n", __FUNCTION__, methods->name));
	return NT_STATUS_NOT_IMPLEMENTED;
}

/***************************************************************
 enumerate all groups
****************************************************************/

static NTSTATUS pgsqlsam_enum_group_mapping(struct pdb_methods *methods,
		enum SID_NAME_USE sid_name_use, GROUP_MAP **rmap, int *num_entries, BOOL unix_only)
{
	DEBUG(10, ("%s(pdbm-%s, type=%d, (out), (out), unix_only=%d)\n", __FUNCTION__, methods->name, sid_name_use, unix_only));
	return NT_STATUS_NOT_IMPLEMENTED;
}

/***************************************************************************
 module init functions
****************************************************************************/

static NTSTATUS pgsqlsam_init(PDB_CONTEXT *pdb_context, PDB_METHODS **pdb_method, const char *location)
{
	NTSTATUS nt_status;
	pgsqlam_private *prv;
	
	if (!NT_STATUS_IS_OK(nt_status = make_pdb_methods(pdb_context->mem_ctx, pdb_method))) {
		return nt_status;
	}
	
	(*pdb_method)->name = "pdb_pgsql";
	
	/* Functions your pdb module doesn't provide should be set 
	 * to NULL */
	
	(*pdb_method)->setsampwent = pgsqlsam_setsampwent;
	(*pdb_method)->endsampwent = pgsqlsam_endsampwent;
	(*pdb_method)->getsampwent = pgsqlsam_getsampwent;
	(*pdb_method)->getsampwnam = pgsqlsam_getsampwnam;
	(*pdb_method)->getsampwsid = pgsqlsam_getsampwsid;
	(*pdb_method)->add_sam_account = pgsqlsam_add_sam_account;
	(*pdb_method)->update_sam_account = pgsqlsam_update_sam_account;
	(*pdb_method)->delete_sam_account = pgsqlsam_delete_sam_account;
	
	(*pdb_method)->getgrsid = pgsqlsam_getgrsid;
	(*pdb_method)->getgrgid = pgsqlsam_getgrgid;
	(*pdb_method)->getgrnam = pgsqlsam_getgrnam;
	(*pdb_method)->add_group_mapping_entry = pgsqlsam_add_group_mapping_entry;
	(*pdb_method)->update_group_mapping_entry = pgsqlsam_update_group_mapping_entry;
	(*pdb_method)->delete_group_mapping_entry = pgsqlsam_delete_group_mapping_entry;
	(*pdb_method)->enum_group_mapping = pgsqlsam_enum_group_mapping;
	
	pgsqlsam_debug_level = debug_add_class("pgsqlsam");
	if (pgsqlsam_debug_level == -1) {
		pgsqlsam_debug_level = DBGC_ALL;
		DEBUG(0, ("pgsqlsam: Couldn't register custom debugging class!\n"));
	} else DEBUG(10, ("pgsqlsam: Debug class number of 'pgsqlsam': %d\n", pgsqlsam_debug_level));
	
	DEBUG(0, ("Initializing pgsqlsam\n"));
	if (location)
		DEBUG(10, ("Location: %s\n", location));
	
	prv = talloc(pdb_context->mem_ctx, sizeof(pgsqlam_private));
	(*pdb_method)->private_data = prv;
	(*pdb_method)->free_private_data = free_private_data;
	prv->conn = NULL;
	prv->location = talloc_strdup(pdb_context->mem_ctx, location);
	prv->domain = talloc_strdup(pdb_context->mem_ctx, get_global_sam_name());
	
#if 0
	{
		int i;
		for (i = 0; i < CONFIG_ALL; i++) {
			DEBUG(10, ("%s = \"%s\" (Default: \"%s\")\n", config_key[i], config_value(prv, i), config_default[i]));
		}
	}
#endif
	DEBUG(10, ("connect info: %s\n", config_value(prv, CONNECTINFO)));
	
	prv->conn = PQconnectdb(config_value(prv, CONNECTINFO));
	if (PQstatus(prv->conn) != CONNECTION_OK) {
		DEBUG(0, ("Failed to connect to postgresql database: error: %s\n", PQerrorMessage(prv->conn)));
		return NT_STATUS_UNSUCCESSFUL;
	}
	
	/*PQtrace(prv->conn, stderr);*/
	
	DEBUG(5, ("Connected to postgresql db\n"));
	return NT_STATUS_OK;
}

NTSTATUS init_module(void)
{
	NTSTATUS nt_status;
	
	if (!NT_STATUS_IS_OK(nt_status = smb_register_passdb(PASSDB_INTERFACE_VERSION, "pdb_pgsql", pgsqlsam_init)))
		return nt_status;
	
	return NT_STATUS_OK;
}
